import param
import plotly.graph_objs as go

from ...core import util
from ...operation import interpolate_curve
from .element import ElementPlot, ColorbarPlot


class ScatterPlot(ColorbarPlot):

    color_index = param.ClassSelector(default=None, class_=(util.basestring, int),
                                      allow_None=True, doc="""
      Index of the dimension from which the color will the drawn""")

    style_opts = ['symbol', 'color', 'cmap', 'fillcolor', 'opacity', 'fill', 'marker', 'size']

    graph_obj = go.Scatter

    def graph_options(self, element, ranges):
        opts = super(ScatterPlot, self).graph_options(element, ranges)
        opts['mode'] = 'markers'
        style = self.style[self.cyclic_index]
        cdim = element.get_dimension(self.color_index)
        if cdim:
            copts = self.get_color_opts(cdim, element, ranges, style)
            copts['color'] = element.dimension_values(cdim)
            opts['marker'] = copts
        else:
            opts['marker'] = style
        return opts


class PointPlot(ScatterPlot):

    def get_data(self, element, ranges):
        data = dict(x=element.dimension_values(0),
                    y=element.dimension_values(1))
        return (), data


class CurvePlot(ElementPlot):

    interpolation = param.ObjectSelector(objects=['linear', 'steps-mid',
                                                  'steps-pre', 'steps-post'],
                                         default='linear', doc="""
        Defines how the samples of the Curve are interpolated,
        default is 'linear', other options include 'steps-mid',
        'steps-pre' and 'steps-post'.""")

    graph_obj = go.Scatter

    style_opts = ['color', 'dash', 'width', 'line_width']

    def graph_options(self, element, ranges):
        if 'steps' in self.interpolation:
            element = interpolate_curve(element, interpolation=self.interpolation)
        opts = super(CurvePlot, self).graph_options(element, ranges)
        opts['mode'] = 'lines'
        style = self.style[self.cyclic_index]
        if 'line_width' in style:
            style['width'] = style.pop('line_width')
        opts['line'] = style
        return opts

    def get_data(self, element, ranges):
        return (), dict(x=element.dimension_values(0),
                        y=element.dimension_values(1))


class ErrorBarsPlot(ElementPlot):

    graph_obj = go.Scatter

    style_opts = ['color', 'dash', 'width', 'opacity', 'thickness']

    def graph_options(self, element, ranges):
        opts = super(ErrorBarsPlot, self).graph_options(element, ranges)
        opts['mode'] = 'lines'
        opts['line'] = {'width': 0}
        return opts

    def get_data(self, element, ranges):
        neg_error = element.dimension_values(2)
        pos_idx = 3 if len(element.dimensions()) > 3 else 2
        pos_error = element.dimension_values(pos_idx)

        style = self.style[self.cyclic_index]
        if 'line_width' in style:
            style['width'] = style.pop('line_width')
        error_y = dict(type='data', array=pos_error,
                       arrayminus=neg_error, visible=True, **style)
        return (), dict(x=element.dimension_values(0),
                        y=element.dimension_values(1),
                        error_y=error_y)


class BivariatePlot(ColorbarPlot):

    ncontours = param.Integer(default=None)
    
    graph_obj = go.Histogram2dcontour

    style_opts = ['cmap']
    
    def graph_options(self, element, ranges):
        opts = super(BivariatePlot, self).graph_options(element, ranges)
        if self.ncontours:
            opts['autocontour'] = False
            opts['ncontours'] = self.ncontours
        style = self.style[self.cyclic_index]
        copts = self.get_color_opts(None, element, ranges, style)
        return dict(opts, **copts)

    def get_data(self, element, ranges):
        return (), dict(x=element.dimension_values(0),
                        y=element.dimension_values(1))


class DistributionPlot(ElementPlot):

    graph_obj = go.Histogram

    def get_data(self, element, ranges):
        return (), dict(x=element.dimension_values(0))


class BarPlot(ElementPlot):

    group_index = param.Integer(default=0, doc="""
       Index of the dimension in the supplied Bars
       Element, which will be laid out into groups.""")

    category_index = param.Integer(default=1, doc="""
       Index of the dimension in the supplied Bars
       Element, which will be laid out into categories.""")

    stack_index = param.Integer(default=2, doc="""
       Index of the dimension in the supplied Bars
       Element, which will stacked.""")

    graph_obj = go.Bar

    def generate_plot(self, key, ranges):
        element = self._get_frame(key)
        ranges = self.compute_ranges(self.hmap, key, ranges)
        ranges = util.match_spec(element, ranges)

        cat_dim = element.get_dimension(self.category_index)
        stack_dim = element.get_dimension(self.stack_index)
        x_dim = element.get_dimension(self.group_index)
        vdim = element.get_dimension(element.ndims)
        if cat_dim and stack_dim:
            self.warning("Plotly does not support stacking and categories "
                         "on a bar chart at the same time.")

        if element.ndims == 1:
            bars = [go.Bar(x=element.dimension_values(x_dim),
                          y=element.dimension_values(vdim))]
        else:
            group_dim = cat_dim if cat_dim else stack_dim
            els = element.groupby(group_dim)
            bars = []
            for k, el in els.items():
                bars.append(go.Bar(x=el.dimension_values(x_dim),
                                   y=el.dimension_values(vdim), name=k))
        layout = self.init_layout(key, element, ranges, x_dim, vdim)
        self.handles['layout'] = layout
        layout['barmode'] = 'group' if cat_dim else 'stacked'

        fig = go.Figure(data=bars, layout=layout)
        self.handles['fig'] = fig
        return fig


class BoxWhiskerPlot(ElementPlot):

    boxpoints = param.ObjectSelector(objects=["all", "outliers",
                                              "suspectedoutliers", False],
                                     default='outliers', doc="""
        Which points to show, valid options are 'all', 'outliers',
        'suspectedoutliers' and False""")

    jitter = param.Number(default=0, doc="""
        Sets the amount of jitter in the sample points drawn. If "0",
        the sample points align along the distribution axis. If "1",
        the sample points are drawn in a random jitter of width equal
        to the width of the box(es).""")

    mean = param.ObjectSelector(default=False, objects=[True, False, 'sd'],
                                doc="""
        If "True", the mean of the box(es)' underlying distribution
        is drawn as a dashed line inside the box(es). If "sd" the
        standard deviation is also drawn.""")

    style_opts = ['color', 'opacity', 'outliercolor', 'symbol',
                  'size']

    def generate_plot(self, key, ranges):
        element = self._get_frame(key)
        ranges = self.compute_ranges(self.hmap, key, ranges)
        ranges = util.match_spec(element, ranges)

        style = self.style[self.cyclic_index]
        orientation = 'h' if self.invert_axes else 'v'
        axis = 'x' if self.invert_axes else 'y'
        box_opts = dict(boxmean=self.mean, jitter=self.jitter,
                        marker=style, orientation=orientation)
        groups = element.groupby(element.kdims)
        groups = groups.data.items() if element.kdims else [(element.label, element)]
        plots = []
        for key, group in groups:
            if element.kdims:
                label = ','.join([d.pprint_value(v) for d, v in zip(element.kdims, key)])
            else:
                label = key
            data = {axis: group.dimension_values(group.vdims[0])}
            plots.append(go.Box(name=label, **dict(box_opts, **data)))
        layout = self.init_layout(key, element, ranges, element.kdims, element.vdims)
        self.handles['layout'] = layout
        fig = go.Figure(data=plots, layout=layout)
        self.handles['fig'] = fig
        return fig

    def get_extents(self, element, ranges):
        return (None, None, None, None)



import numpy as np
import plotly.graph_objs as go
import param

from ...core.util import basestring
from .plot import PlotlyPlot
from ..plot import GenericElementPlot, GenericOverlayPlot
from .. import util


class ElementPlot(PlotlyPlot, GenericElementPlot):

    aspect = param.Parameter(default='cube', doc="""
        The aspect ratio mode of the plot. By default, a plot may
        select its own appropriate aspect ratio but sometimes it may
        be necessary to force a square aspect ratio (e.g. to display
        the plot as an element of a grid). The modes 'auto' and
        'equal' correspond to the axis modes of the same name in
        matplotlib, a numeric value may also be passed.""")

    bgcolor = param.ClassSelector(class_=(str, tuple), default=None, doc="""
        If set bgcolor overrides the background color of the axis.""")

    invert_axes = param.ObjectSelector(default=False, doc="""
        Inverts the axes of the plot. Note that this parameter may not
        always be respected by all plots but should be respected by
        adjoined plots when appropriate.""")

    invert_xaxis = param.Boolean(default=False, doc="""
        Whether to invert the plot x-axis.""")

    invert_yaxis = param.Boolean(default=False, doc="""
        Whether to invert the plot y-axis.""")

    invert_zaxis = param.Boolean(default=False, doc="""
        Whether to invert the plot z-axis.""")

    labelled = param.List(default=['x', 'y'], doc="""
        Whether to plot the 'x' and 'y' labels.""")

    logx = param.Boolean(default=False, doc="""
         Whether to apply log scaling to the x-axis of the Chart.""")

    logy  = param.Boolean(default=False, doc="""
         Whether to apply log scaling to the y-axis of the Chart.""")

    logz  = param.Boolean(default=False, doc="""
         Whether to apply log scaling to the y-axis of the Chart.""")

    margins = param.NumericTuple(default=(50, 50, 50, 50), doc="""
         Margins in pixel values specified as a tuple of the form
         (left, bottom, right, top).""")

    show_legend = param.Boolean(default=False, doc="""
        Whether to show legend for the plot.""")

    xaxis = param.ObjectSelector(default='bottom',
                                 objects=['top', 'bottom', 'bare', 'top-bare',
                                          'bottom-bare', None], doc="""
        Whether and where to display the xaxis, bare options allow suppressing
        all axis labels including ticks and xlabel. Valid options are 'top',
        'bottom', 'bare', 'top-bare' and 'bottom-bare'.""")

    xticks = param.Parameter(default=None, doc="""
        Ticks along x-axis specified as an integer, explicit list of
        tick locations, list of tuples containing the locations and
        labels or a matplotlib tick locator object. If set to None
        default matplotlib ticking behavior is applied.""")

    yaxis = param.ObjectSelector(default='left',
                                      objects=['left', 'right', 'bare', 'left-bare',
                                               'right-bare', None], doc="""
        Whether and where to display the yaxis, bare options allow suppressing
        all axis labels including ticks and ylabel. Valid options are 'left',
        'right', 'bare' 'left-bare' and 'right-bare'.""")

    yticks = param.Parameter(default=None, doc="""
        Ticks along y-axis specified as an integer, explicit list of
        tick locations, list of tuples containing the locations and
        labels or a matplotlib tick locator object. If set to None
        default matplotlib ticking behavior is applied.""")

    graph_obj = None

    def initialize_plot(self, ranges=None):
        """
        Initializes a new plot object with the last available frame.
        """
        # Get element key and ranges for frame
        fig = self.generate_plot(self.keys[-1], ranges)
        self.drawn = True
        return fig


    def generate_plot(self, key, ranges):
        element = self._get_frame(key)
        if element is None:
            return self.handles['fig']
        plot_opts = self.lookup_options(element, 'plot').options
        self.set_param(**{k: v for k, v in plot_opts.items()
                          if k in self.params()})
        self.style = self.lookup_options(element, 'style')

        ranges = self.compute_ranges(self.hmap, key, ranges)
        ranges = util.match_spec(element, ranges)

        data_args, data_kwargs = self.get_data(element, ranges)
        opts = self.graph_options(element, ranges)
        graph = self.init_graph(data_args, dict(opts, **data_kwargs))
        self.handles['graph'] = graph

        layout = self.init_layout(key, element, ranges)
        self.handles['layout'] = layout

        if isinstance(graph, go.Figure):
            graph.update({'layout': layout})
            self.handles['fig'] = graph
        else:
            if not isinstance(graph, list):
                graph = [graph]
            fig = go.Figure(data=graph, layout=layout)
            self.handles['fig'] = fig
            return fig


    def graph_options(self, element, ranges):
        if self.overlay_dims:
            legend = ', '.join([d.pprint_value_string(v) for d, v in
                                self.overlay_dims.items()])
        else:
            legend = element.label

        opts = dict(showlegend=self.show_legend,
                    legendgroup=element.group,
                    name=legend)

        if self.layout_num:
            opts['xaxis'] = 'x' + str(self.layout_num)
            opts['yaxis'] = 'y' + str(self.layout_num)

        return opts


    def init_graph(self, plot_args, plot_kwargs):
        return self.graph_obj(*plot_args, **plot_kwargs)


    def get_data(self, element, ranges):
        return {}


    def init_layout(self, key, element, ranges, xdim=None, ydim=None):
        l, b, r, t = self.get_extents(element, ranges)

        options = {}

        xdim = element.get_dimension(0) if xdim is None else xdim
        ydim = element.get_dimension(1) if ydim is None else ydim
        xlabel, ylabel, zlabel = self._get_axis_labels([xdim, ydim])

        if self.invert_axes:
            xlabel, ylabel = ylabel, xlabel
            l, b, r, t = b, l, t, r

        if 'x' not in self.labelled:
            xlabel = ''
        if 'y' not in self.labelled:
            ylabel = ''

        if xdim:
            xaxis = dict(range=[l, r], title=xlabel)
            if self.logx:
                xaxis['type'] = 'log'
            options['xaxis'] = xaxis

        if ydim:
            yaxis = dict(range=[b, t], title=ylabel)
            if self.logy:
                yaxis['type'] = 'log'
            options['yaxis'] = yaxis

        l, b, r, t = self.margins
        margin = go.Margin(l=l, r=r,b=b, t=t, pad=4)
        return go.Layout(width=self.width, height=self.height,
                         title=self._format_title(key, separator=' '),
                         plot_bgcolor=self.bgcolor, margin=margin,
                         **options)


    def update_frame(self, key, ranges=None):
        """
        Updates an existing plot with data corresponding
        to the key.
        """
        self.generate_plot(key, ranges)


class ColorbarPlot(ElementPlot):

    colorbar = param.Boolean(default=False, doc="""
        Whether to display a colorbar.""")

    colorbar_opts = param.Dict(default={}, doc="""
        Allows setting including borderwidth, showexponent, nticks,
        outlinecolor, thickness, bgcolor, outlinewidth, bordercolor,
        ticklen, xpad, ypad, tickangle...""")

    def get_color_opts(self, dim, element, ranges, style):
        opts = {}
        if self.colorbar:
            opts['colorbar'] = dict(title=dim.pprint_label,
                                    **self.colorbar_opts)
        else:
            opts['showscale'] = False

        cmap = style.pop('cmap', 'viridis')
        if cmap == 'fire':
            values = np.linspace(0, 1, len(util.fire_colors))
            cmap = [(v, 'rgb(%d, %d, %d)' % tuple(c))
                    for v, c in zip(values, np.array(util.fire_colors)*255)]
        elif isinstance(cmap, basestring):
            if cmap[0] == cmap[0].lower():
                cmap = cmap[0].upper() + cmap[1:]
            if cmap.endswith('_r'):
                cmap = cmap[:-2]
                opts['reversescale'] = True
        opts['colorscale'] = cmap
        if dim:
            cmin, cmax = ranges.get(dim.name, element.range(dim.name))
            opts['cmin'] = cmin
            opts['cmax'] = cmax
            opts['cauto'] = False
        return dict(style, **opts)


class OverlayPlot(GenericOverlayPlot, ElementPlot):


    def initialize_plot(self, ranges=None):
        """
        Initializes a new plot object with the last available frame.
        """
        # Get element key and ranges for frame
        return self.generate_plot(list(self.hmap.data.keys())[0], ranges)


    def generate_plot(self, key, ranges):
        element = self._get_frame(key)

        ranges = self.compute_ranges(self.hmap, key, ranges)
        figure = None
        for okey, subplot in self.subplots.items():
            fig = subplot.generate_plot(key, ranges)
            if figure is None:
                figure = fig
            else:
                figure['data'].extend(fig['data'])

        layout = self.init_layout(key, element, ranges)
        figure['layout'].update(layout)
        self.handles['fig'] = figure
        return figure
